import os
import dlchord
from dlchord import Chord
from pywfd import label as lb

chord_quality = {
    0: "",
    1: "m",
    2: "dim7",
    3: "aug",
    4: "6",
    5: "7",
    6: "M7",
    7: "add9",
    8: "m6",
    9: "m7",
    10: "mM7",
    11: "sus4",
    12: "7sus4",
    13: "m7-5"}
chord_tone = {
    0: "C",
    1: "C#",
    2: "D",
    3: "D#",
    4: "E",
    5: "F",
    6: "F#",
    7: "G",
    8: "G#",
    9: "A",
    10: "A#",
    11: "B"}

tones = ["C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B", "N"]


def splitindex(l, n):
    for idx in range(0, len(l), n):
        yield l[idx:idx + n]


def number_to_chord(chord_num, is_input=False, on_chord=False):
    pitch = (chord_num % 12)
    name = 0

    if on_chord:
        pitch -= 1
    else:
        name = chord_num // 12

    if is_input:
        name = (chord_num - 11) // 12
        pitch += 1

    pitch %= 12
    name %= 14

    return chord_tone[pitch] + chord_quality[name]


def chord_to_array(chord, tension=True):
    chord_array = [0] * 48
    if chord == '':
        return chord_array

    chord_array[0] = 1

    if chord == "N.C.":
        return chord_array

    chord_num = 11

    try:
        chord = Chord(chord)
    except ValueError:
        return chord_array

    quality = chord.quality.quality

    if tension:
        if chord.bass == chord.root:
            if "b9" in quality:
                if "m" in quality:
                    chord = Chord(Chord(tones[chord.root] + "7").transpose(3).chord + "/" + tones[chord.root])
                else:
                    chord = Chord(Chord(tones[chord.root] + "dim7").transpose(4).chord + "/" + tones[chord.root])
            elif "9" in quality and "add" not in quality and "sus" not in quality and "#9" not in quality:
                if "m" in quality and "M" not in quality:
                    chord = Chord(Chord(tones[chord.root] + "M7").transpose(3).chord + "/" + tones[chord.root])
                elif "M" in quality:
                    chord = Chord(Chord(tones[chord.root] + "m7").transpose(4).chord + "/" + tones[chord.root])
                else:
                    chord = Chord(Chord(tones[chord.root] + "m7-5").transpose(4).chord + "/" + tones[chord.root])
            elif "11" in quality:
                if "m" in quality:
                    chord = Chord(Chord(tones[chord.root] + "add9").transpose(3).chord + "/" + tones[chord.root])

    quality = chord.quality.quality
    chord_num += chord.root

    sorted_quality = sorted(chord_quality.items(), key=lambda x: len(x[1]), reverse=True)
    for idx, qua in sorted_quality:
        if qua[:3] in quality:
            chord_num += (12 * idx)
            break
    
    chord_array[4] = chord_num
    if chord.bass != chord.root:
        chord_array[5] = chord.bass + 1

    return chord_array


def chord_label(chord_time):
    result_text = ""

    for times in chord_time:
        result_text += str(times[0]) + ":" + str(times[1]) + ":" + times[2] + os.linesep

    return result_text


def label_to_chord(label, sep=':'):
    times = []

    for frame in label:
        if frame:
            sp = frame.strip().split(sep)
            if len(sp) == 2:
                chord = "N.C."
            else:
                chord = str(sp[2])
            
            times.append([float(sp[0]), float(sp[1]), chord])

    return times


def is_wavetone_chord(chord):
    tension = ["9", "13", "11", "M7-5", "aug7", "augM7"]
    if any([t in chord for t in tension]):
        return False
    else:
        return True


class ChordSplit:
    def __init__(self, chord, rhythm, label):
        self._raw_chord = chord
        self.rhythm = rhythm
        self.label = label
        self._chord = self._raw_chord_to_chord(self._raw_chord)
        self.tempo = self.rhythm.tempo(0)
        self.beat_offset = self.rhythm.beat_offset / 1000

    @property
    def raw_chord(self):
        return self._raw_chord

    @property
    def chord(self):
        return self._chord

    def _raw_chord_to_chord(self, raw_chord):
        chord = []
        for i in range(len(raw_chord)):
            parse_chord = self._split(i)
            if parse_chord != "" and parse_chord != "N.C." and self.rhythm.musickey(i) != 12:
                parse_chord = dlchord.Chord(parse_chord).modify(key=tones[self.rhythm.musickey(i)]).chord
            chord.append(parse_chord)

        return chord

    def chordLabel(self):
        """[summary]
        
        Returns:
            {x: [start_time, end_time, chord]}
        """
        s_time = 0.0 + self.beat_offset
        e_time = 0.0
        times = []

        chord = ""
        now_chord = ""
        for i in range(len(self.chord)):
            now_chord = self.chord[i]

            if i == 0:
                chord = now_chord

            if chord != now_chord and now_chord != '' or i == (len(self.chord) - 1):
                e_time = self.rhythm.time(i)
                times.append([s_time, e_time, chord])
                chord = now_chord
                s_time = e_time

        return times

    def keyLabel(self):
        times = []

        s_time = 0.0 + self.beat_offset
        e_time = 0.0
        key = ""
        now_key = ""
        
        for i, frame in enumerate(self.rhythm.rhythmkey.rhythm_key_map):
            now_key = frame[-1]
            
            if i == 0:
                key = now_key

            if key != now_key or i == (len(self.rhythm.rhythmkey.rhythm_key_map) - 1):
                e_time = self.rhythm.time(i)
                times.append([s_time, e_time, tones[key]])
                key = now_key
                s_time = e_time

        return times

    def label_to_array(self, label):
        chordlist = []
        offset = 0

        min_time = self.rhythm.time(0)
        duration = abs(min_time - self.rhythm.time(1))

        for i in range(len(label)):
            if label[i][1] < min_time or abs(label[i][1] - min_time) < duration:
                offset += 1

        label = label[offset:]
        offset = 0

        if min_time - label[0][0] > duration:
            label[0][0] += abs(min_time - label[0][0])

        for i in range(len(self.chord)):
            if len(label) <= offset:
                break
            
            duration = abs(self.rhythm.time(i) - self.rhythm.time(i + 1))

            dist = abs(self.rhythm.time(i) - label[offset][0])
            if dist < duration:
                if dist <= abs(self.rhythm.time(i + 1) - label[offset][0]):
                    chordlist.append(chord_to_array(label[offset][-1]))
                    if not is_wavetone_chord(label[offset][-1]):
                        c_label = lb.LabelSplit()
                        c_label.setTime(int(self.rhythm.time(i) * 1000))
                        c_label.setLabel(label[offset][-1])
                        self.label.append(c_label)
                    offset += 1
                else:
                    chordlist.append(chord_to_array(''))
            else:
                chordlist.append(chord_to_array(''))

        return chordlist

    def to_chordpro(self, indent=4, form=True, advanced=False):
        chordtext = ""
        measure = -1
        time = 0
        measure_count = 0

        for i, chord in enumerate(self.chord, start=1):
            if form:
                try:
                    if self.rhythm.musickey(i - 1) != 12:
                        chord = dlchord.Chord(chord).modify(key=tones[self.rhythm.musickey(i - 1)], advanced=advanced)
                except ValueError:
                    pass
                
            measure_num = self.rhythm.measure_number(i - 1)
            if measure != self.rhythm.measure(measure_num):
                measure = self.rhythm.measure(measure_num)
                time = i - 1

            if chord:
                chordtext += "[{}]-".format(chord)
            else:
                chordtext += "-"

            if (i - time) % (measure * 2) == 0:
                chordtext += "|"
                measure_count += 1

            elif (i - time) % measure == 0:
                chordtext += " "
            
            if measure_count % indent == 0 and measure_count != 0:
                measure_count = 0
                chordtext += os.linesep

        return chordtext

    def _split(self, time):
        enable_chord = self.raw_chord[time][0]
        input_chord = self.raw_chord[time][4]
        input_on_chord = self.raw_chord[time][5]
        auto_chord = self.raw_chord[time][8]

        result_chord = ""
        # コードが存在していたら
        if enable_chord:
            # コード入力の場合
            if input_chord > 10:
                result_chord = number_to_chord(input_chord, is_input=True)
                if input_on_chord:
                    result_chord += "/" + \
                        number_to_chord(input_on_chord, on_chord=True)
            # N.C.の場合
            elif input_chord == 0:
                result_chord = "N.C."

            # 自動認識の場合
            else:
                auto_chord = self.raw_chord[time][4 + (4 * input_chord)]
                result_chord = number_to_chord(auto_chord)
        
        return result_chord



class RhythmKey:
    def __init__(self, rhythm_key_map, chord):
        self.raw_rhythm_key_map = rhythm_key_map
        self.rhythm_key_map = self._convert_rhythm_key_map(
            self.raw_rhythm_key_map, chord)
        self.chord = chord

    def _convert_rhythm_key_map(self, raw_rhythm_key_map, chord):
        rhythm_key_map = []

        for i in range(len(raw_rhythm_key_map)):
            if i == len(raw_rhythm_key_map) - 1:
                rmp = raw_rhythm_key_map[i][1:]
                if raw_rhythm_key_map[i][-1] > 12:
                    rmp[-1] = 12
                loop = int(len(chord) - len(rhythm_key_map))
                for j in range(loop):
                    rhythm_key_map.append(rmp)
                break

            loop = int(raw_rhythm_key_map[i + 1][0] - raw_rhythm_key_map[i][0]) * raw_rhythm_key_map[i][1] * 2
            rmp = raw_rhythm_key_map[i][1:]
            if raw_rhythm_key_map[i][-1] > 12:
                rmp[-1] = 12
            for j in range(loop):
                
                rhythm_key_map.append(rmp)

        return rhythm_key_map


class TempoMap:

    def __init__(self, tempomap, rhythmkey, beat_offset, k=960):
        self.raw_tempomap = tempomap
        self.rhythm_key_map = rhythmkey.rhythm_key_map

        self.beat_offset = beat_offset
        self.k = k

        self.tempomap = self._convert_tempomap(self.raw_tempomap)

    def _convert_tempomap(self, raw_tempomap):
        tempomap = []

        nowtempo = raw_tempomap[0][1] / 10000
        if len(raw_tempomap) > 1:
            now_time = 1
        else:
            now_time = 0

        for i in range(len(self.rhythm_key_map)):
            tempomap.append(nowtempo)
            
            if raw_tempomap[now_time][0] / self.k == i:
                nowtempo = raw_tempomap[now_time][1] / 10000
                if now_time < len(raw_tempomap) - 1:
                    now_time += 1

        return tempomap


class Rhythm:
    def __init__(self, tempomap, rhythmkey):
        self.tempomap = tempomap
        self.rhythmkey = rhythmkey
        self._length = self.time(len(self.tempomap.tempomap))

    @property
    def beat_offset(self):
        return self.tempomap.beat_offset

    @property
    def length(self):
        return self._length

    def time(self, index):
        times = 0.0

        for tempo in self.tempomap.tempomap[:index]:
            times += (1 / ((tempo / 60) * 2))

        return times + (self.beat_offset / 1000)

    def musickey(self, index):
        return self.rhythmkey.rhythm_key_map[index][1]

    def tempo(self, index):
        return self.tempomap.tempomap[index]

    def measure(self, index):
        return self.rhythmkey.rhythm_key_map[index][0]

    def measure_number(self, index):
        number = 0
        measure_sum = 0

        for i in range(len(self.rhythmkey.rhythm_key_map)):
            measure_sum += 1

            if measure_sum > index:
                break

            number += 1

        return number



from . import wfd
from . import io
from . import chord

from .chord import number_to_chord
from .chord import chord_label, label_to_chord
from .rhythm import RhythmKey, TempoMap
from .label import Label, LabelSplit
from .wfd import load
from .wfd import write



from pywfd import chord
from pywfd import rhythm
from pywfd import io
from pywfd import label as lb
import numpy as np


class WFDData:
    def __init__(self, loader):
        self._loader = loader
        self._loader.readHeader()
        self._loader.readIndex()
        self._loader.readData()

        self.tempo = self._loader.headers[self._loader.headerA(
            lb.TEMPO, method="DATATYPE")].value
        self.block_per_semitone = self._loader.headers[self._loader.headerA(
            lb.BLOCK_PER_SEMITONE, method="DATATYPE")].value
        self.min_note = self._loader.headers[self._loader.headerA(
            lb.MIN_NOTE, method="DATATYPE")].value
        self.range_of_scale = self._loader.headers[self._loader.headerA(
            lb.RANGE_OF_SCALE, method="DATATYPE")].value
        self.block_per_second = self._loader.headers[self._loader.headerA(
            lb.BLOCK_PER_SECOND, method="DATATYPE")].value
        self.time_all_block = self._loader.headers[self._loader.headerA(
            lb.TIME_ALL_BLOCK, method="DATATYPE")].value
        self.beat_offset = self._loader.headers[self._loader.headerA(lb.OFFSET, method="DATATYPE")].value
        self.beat = self._loader.headers[self._loader.headerA(
            lb.BEAT, method="DATATYPE")].value

        self.label_list = lb.Label(self.getdata(lb.LABEL_LIST))
        self._rhythmkey = rhythm.RhythmKey(
            self.getdata(
                lb.RHYTHM_KEYMAP),
            chord=self.getdata(
                lb.CHORD_RESULT))

        self._tempomap = rhythm.TempoMap(
            self.getdata(
                lb.TEMPO_MAP),
            rhythmkey=self.rhythmkey,
            beat_offset=self.beat_offset)

        self._chords = chord.ChordSplit(
            self.getdata(
                lb.CHORD_RESULT),
            rhythm=rhythm.Rhythm(self.tempomap, self.rhythmkey),
            label=self.label_list)

    @property
    def loader(self):
        return self._loader

    @property
    def spectrumStereo(self):
        """音声スペクトル(stereo)"""
        return self.getdata(lb.SPECTRUM_STEREO)

    @spectrumStereo.setter
    def spectrumStereo(self, spectrum):
        spectrum = np.array(spectrum).flatten()
        self.setdata(lb.SPECTRUM_STEREO, spectrum)

    @property
    def spectrumLRM(self):
        """音声スペクトル(L-R)"""
        return self.getdata(lb.SPECTRUM_LR_M)

    @spectrumLRM.setter
    def spectrumLRM(self, spectrum):
        spectrum = np.array(spectrum).flatten()
        self.setdata(lb.SPECTRUM_LR_M, spectrum)

    @property
    def spectrumLRP(self):
        """音声スペクトル(L+R)"""
        return self.getdata(lb.SPECTRUM_LR_P)

    @spectrumLRP.setter
    def spectrumLRP(self, spectrum):
        spectrum = np.array(spectrum).flatten()
        self.setdata(lb.SPECTRUM_LR_P, spectrum)

    @property
    def spectrumL(self):
        """音声スペクトル(L)"""
        return self.getdata(lb.SPECTRUM_L)

    @spectrumL.setter
    def spectrumL(self, spectrum):
        spectrum = np.array(spectrum).flatten()
        self.setdata(lb.SPECTRUM_L, spectrum)

    @property
    def spectrumR(self):
        """	音声スペクトル(R)"""
        return self.getdata(lb.SPECTRUM_R)

    @spectrumR.setter
    def spectrumR(self, spectrum):
        spectrum = np.array(spectrum).flatten()
        self.setdata(lb.SPECTRUM_R, spectrum)

    @property
    def chords(self):
        return self._chords

    @chords.setter
    def chords(self, chords):
        verify = np.array(self.chords_raw)
        chords = np.concatenate((verify[:16], np.array(chords).flatten(), verify[-48:])).astype("uint8")
        self.setdata(lb.CHORD_RESULT, chords)
        label = np.array(self.chords.label.to_array()).astype("uint32")
        self.setdata(lb.LABEL_LIST, label)
        self._loader.shift_offset(lb.LABEL_LIST, int(len(label) * 4))

    @property
    def rhythmkey(self):
        return self._rhythmkey

    @property
    def tempomap(self):
        return self._tempomap

    @property
    def chords_raw(self):
        return self.get_raw_data(lb.CHORD_RESULT)

    @chords_raw.setter
    def chords_raw(self, chords):
        self.setdata(lb.CHORD_RESULT, chords)

    def setdata(self, key, data):
        self._loader.raw_data[key] = data

    def getdata(self, key):
        return self._loader.wfd_data[key]

    def get_raw_data(self, key):
        return self._loader.raw_data[key]


class WFD:
    def __init__(self):
        self._loader = io.WFDLoader()

    def load(self, filepath):
        self._loader.open(filepath)
        self.wfd_data = WFDData(self._loader)
        return self.wfd_data


def load(file):
    wfd = WFD()

    data = wfd.load(file)

    return data


def write(file, data):
    io.WFDWriter().write(file, data)



import numpy as np
import struct
import os.path
import sys

from pywfd import label as lb
from pywfd import chord as cp


class Writer:
    def __init__(self):
        pass
    
    def write(self, file, wfd_data):
        pass


class Loader:
    def __init__(self):
        self._buffer = None
        self.offset = 0
        self.offset_list = []

    @property
    def buffer(self):
        return self._buffer

    def open(self, filepath):
        self._buffer = open(filepath, 'rb').read()

    def unpack(self, buffer, format, count, offset, add_offset=True):
        self.offset_list.append(self.offset)
        data = np.frombuffer(buffer, dtype=format, count=count, offset=offset)
        self.offset += int(struct.calcsize(format) * count) if add_offset else 0

        return data


class WFDWriter(Writer):
    def __init__(self):
        super().__init__()

    def parse_wfd(self, wfd_data):
        header_offset, index_offset, data_offset = wfd_data.loader.info()
        result = {}
        for k, v in header_offset.items():
            header = wfd_data.loader.headers[wfd_data.loader.headerA(k, "DATATYPE")]
            result[v[0]] = [header.value, v[1]]

        for k, v in index_offset.items():
            for v_list in v:
                index = wfd_data.loader.indexes[wfd_data.loader.indexA(k, "DATATYPE")]
                result[v_list[0]] = [index.search(v_list[2]), v_list[1]]

        for k, v in data_offset.items():
            result[v[0]] = [wfd_data.get_raw_data(k), v[1]]
        return result

    def write(self, file, wfd_data):
        write_data = sorted(self.parse_wfd(wfd_data).items(), key=lambda x: x[0])
        with open(file, 'wb') as f:
            for offset, value in write_data:
                f.seek(offset)
                if type(value[0]) is np.ndarray:
                    value[0] = bytearray(value[0])
                    f.write(value[0])
                else:
                    value[0] = int(value[0])

                    f.write(
                        value[0].to_bytes(
                            struct.calcsize(
                                value[1]),
                            byteorder=sys.byteorder))


class WFDHeader:
    def __init__(self, datatype, datanum, value):
        self._datatype = datatype
        self._datanum = datanum
        self._value = value

    @property
    def datatype(self):
        return self._datatype

    @property
    def datanum(self):
        return self._datanum

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


class WFDIndex:
    def __init__(self, datatype, datanum, datasize, dataformat, index):
        self._datatype = datatype
        self._datanum = datanum
        self._datasize = datasize
        self._dataformat = dataformat
        self._index = index

    @property
    def datatype(self):
        return self._datatype

    @property
    def datanum(self):
        return self._datanum

    @property
    def datasize(self):
        return self._datasize

    @datasize.setter
    def datasize(self, datasize):
        self._datasize = datasize

    @property
    def dataformat(self):
        return self._dataformat

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, index):
        self._index = index

    def search(self, key):
        if key == "DATATYPE":
            return self._datatype
        elif key == "DATANUM":
            return self._datanum
        elif key == "DATASIZE":
            return self._datasize
        elif key == "DATAFORMAT":
            return self._dataformat
        elif key == "INDEX":
            return self._index

        
class WFDLoader(Loader):
    def __init__(self):
        super().__init__()
        self.headers = [
            WFDHeader(lb.FILETYPE, 0, 0),
            WFDHeader(lb.RESERVE_SPACE1, 1, 0),
            WFDHeader(lb.RESERVE_SPACE2, 2, 0),
            WFDHeader(lb.BLOCK_PER_SEMITONE, 3, 0),
            WFDHeader(lb.MIN_NOTE, 4, 0),
            WFDHeader(lb.RANGE_OF_SCALE, 5, 0),
            WFDHeader(lb.BLOCK_PER_SECOND, 6, 0),
            WFDHeader(lb.RESERVE_SPACE3, 7, 0),
            WFDHeader(lb.TIME_ALL_BLOCK, 8, 0),
            WFDHeader(lb.BITS_OF_GRAPH, 9, 0),
            WFDHeader(lb.BEAT_DISPLAY_FLAG, 10, 0),
            WFDHeader(lb.TEMPO, 11, 0),
            WFDHeader(lb.OFFSET, 12, 0),
            WFDHeader(lb.BEAT, 13, 0)
        ]
        self.indexes = [
            WFDIndex(lb.DATASIZE, -1, 0, "I", -1),
            WFDIndex(lb._, 0, 0, "H", 0),
            WFDIndex(lb.TEMPO_RESULT, 2, 0, "I", 0),
            WFDIndex(lb.EXTEND_INFO, 4, 0, "I", 0),
            WFDIndex(lb.LABEL_LIST, 6, 0, "I", 0),
            WFDIndex(lb.SPECTRUM_STEREO, 7, 0, "H", 0),
            WFDIndex(lb.SPECTRUM_LR_M, 8, 0, "H", 0),
            WFDIndex(lb.SPECTRUM_LR_P, 9, 0, "H", 0),
            WFDIndex(lb.SPECTRUM_L, 10, 0, "H", 0),
            WFDIndex(lb.SPECTRUM_R, 11, 0, "H", 0),
            WFDIndex(lb.TEMPO_MAP, 12, 0, "I", 0),
            WFDIndex(lb.CHORD_RESULT, 14, 0, "B", 0),
            WFDIndex(lb.RHYTHM_KEYMAP, 15, 0, "I", 0),
            WFDIndex(lb.NOTE_LIST, 16, 0, "I", 0),
            WFDIndex(lb.TEMPO_VOLUME, 17, 0, "I", 0),
            WFDIndex(lb.FREQUENCY, 18, 0, "I", 0),
            WFDIndex(lb.TRACK_SETTING, 19, 0, "I", 0)
        ]

        self.headers = sorted(self.headers, key=lambda x: x.datanum)
        self.indexes = sorted(self.indexes, key=lambda x: x.datanum)

        self.wfd_data = {}

        self.header_offset = {}
        self.index_offset = {}
        self.data_offset = {}

        self.header_format = "I"
        self.index_format = "I"

    @property
    def __indexes__(self):
        return self.indexes

    def open(self, filepath):
        _, ext = os.path.splitext(filepath)
        if ext.lower() != ".wfd":
            raise ValueError("wfdファイルではありません")

        self._buffer = open(filepath, 'rb').read()
        
    def info(self):
        return self.header_offset, self.index_offset, self.data_offset

    def shift_offset(self, datatype, shift_value):
        shift = self.data_offset[datatype][0]
        data = self.indexes[self.indexA(datatype, "DATATYPE")]
        datasize = data.datasize
        data.datasize = shift_value
        shift_value = abs(datasize - shift_value)
        data_offset = {}
        for dtype, value in self.data_offset.items():
            offset = value[0]
            if offset > shift:
                value[0] += shift_value
            data_offset[dtype] = value

        self.data_offset = data_offset

    def readHeader(self):
        """Headerを読み込みます"""
        
        for i in range(len(self.headers)):
            self.header_offset[self.headers[self.headerA(i)].datatype] = [self.offset, self.header_format]
            data = self.unpack(self.buffer, self.header_format, 1, self.offset)
            self.headers[self.headerA(i)].value = data[0]

        return self.headers

    def readIndex(self):
        """Indexを読み込みます"""
        if self.offset >= (struct.calcsize(self.header_format) * len(self.headers)):
            counter = 1
            start_index = self.indexA(-1)
            for i in range(len(self.indexes)):
                if self.indexes[i].datanum == -1:
                    offset = self.offset
                    data = self.unpack(self.buffer, self.index_format, 1, self.offset)
                    self.indexes[start_index].datasize = data[0]
                    self.index_offset[self.indexes[start_index].datatype] = [
                        [offset, self.index_format, "DATASIZE"]]
                else:
                    if self.indexes[start_index].datasize < counter:
                        continue
                    
                    datanumber_offset = self.offset
                    datanumber = self.unpack(self.buffer, self.index_format, 1, self.offset)[0]
                    datanumber = self.indexA(datanumber)

                    datasize_offset = self.offset
                    datasize = self.unpack(self.buffer, self.index_format, 1, self.offset)[0]

                    self.index_offset[self.indexes[datanumber].datatype] = [
                        [datanumber_offset, self.index_format, "DATANUM"],
                        [datasize_offset, self.index_format, "DATASIZE"]
                    ]

                    self.indexes[datanumber].datasize = datasize
                    self.indexes[datanumber].index = counter
                    
                    counter += 1
            
            self.indexes = sorted(self.indexes, key=lambda x: x.index)
        return self.indexes

    def readData(self):
        """データを読み込みます"""
        bps = self.headers[self.headerA(lb.BLOCK_PER_SEMITONE, "DATATYPE")].value
        range_scale = self.headers[self.headerA(lb.RANGE_OF_SCALE, "DATATYPE")].value
        time_all_block = self.headers[self.headerA(lb.TIME_ALL_BLOCK, "DATATYPE")].value
        freq_all_block = bps * range_scale
        data = {}

        for i in range(len(self.indexes)):
            dtype = self.indexes[i]._datatype
            if self.indexes[i].index <= 0:
                data[dtype] = []
                continue

            self.data_offset[dtype] = [self.offset, self.indexes[i].dataformat]
            datasize = self.indexes[i].datasize
            dataformat = self.indexes[i].dataformat
                
            data[dtype] = self.unpack(self.buffer, dataformat, int(datasize / struct.calcsize(dataformat)), self.offset)

        result_data = {}
        for k, v in data.items():
            if len(v) == (time_all_block * freq_all_block):
                result_data[k] = np.array(self.__spectrumData(v, time_all_block, freq_all_block), dtype="float32").T
            elif k == lb.CHORD_RESULT:
                v = v[16:]
                result_data[k] = np.array(list(cp.splitindex(v, 48))[:-1])
            elif k == lb.RHYTHM_KEYMAP:
                result_data[k] = np.array(list(cp.splitindex(v, 3)))
            elif k == lb.TEMPO_MAP:
                result_data[k] = np.array(list(cp.splitindex(v, 2)))
            else:
                result_data[k] = v
        self.wfd_data = result_data
        self.raw_data = data
        return result_data

    def __spectrumData(self, x, time_all_block, freq_all_block):
        """正規化とリシェイプを行います"""
        data = np.array(x / 65535.0, dtype="float32")
        return np.reshape(data, (time_all_block, freq_all_block))

    def headerA(self, key, method="DATANUM"):
        for i in range(len(self.headers)):
            if method == "DATANUM":
                if self.headers[i].datanum == key:
                    return i
            elif method == "DATATYPE":
                if self.headers[i].datatype == key:
                    return i

    def indexA(self, key, method="DATANUM"):
        for i in range(len(self.indexes)):
            if method == "DATANUM":
                if self.indexes[i].datanum == key:
                    return i
            elif method == "DATATYPE":
                if self.indexes[i].datatype == key:
                    return i



from setuptools import setup, find_packages

setup(
    name="pywfd",
    version="1.2.3",
    description="A library to handle wfd in python.",
    install_requires=["numpy", "dlchord"],
    author="anime-song",
    url="https://github.com/anime-song/wfdload",
    license="MIT",
    packages=["pywfd"],
)



import numpy as np
DATASIZE = "datasize"
_ = "_"
TEMPO_RESULT = "tempo_result"
EXTEND_INFO = "extend_info"
LABEL_LIST = "label_list"
SPECTRUM_STEREO = "spectrum_stereo"
SPECTRUM_LR_M = "spectrum_LR_M"
SPECTRUM_LR_P = "spectrum_LR_P"
SPECTRUM_L = "spectrum_L"
SPECTRUM_R = "spectrum_R"
TEMPO_MAP = "tempo_map"
CHORD_RESULT = "chord_result"
RHYTHM_KEYMAP = "rhythm_keymap"
NOTE_LIST = "note_list"
TEMPO_VOLUME = "tempo_volume"
FREQUENCY = "frequency"
TRACK_SETTING = "track_setting"

FILETYPE = "filetype"
RESERVE_SPACE1 = "reserve_space1"
RESERVE_SPACE2 = "reserve_space2"
BLOCK_PER_SEMITONE = "block_per_semitone"
MIN_NOTE = "min_note"
RANGE_OF_SCALE = "range_of_scale"
BLOCK_PER_SECOND = "block_per_second"
RESERVE_SPACE3 = "reserve_space3"
TIME_ALL_BLOCK = "time_all_block"
BITS_OF_GRAPH = "bits_of_graph"
BEAT_DISPLAY_FLAG = "beat_display_flag"
TEMPO = "tempo"
OFFSET = "offset"
BEAT = "beat"


def splitindex(l, n):
    for idx in range(0, len(l), n):
        yield l[idx:idx + n]


class Label:

    def __init__(self, label):
        self._raw_label = label
        self._label_list = []
        self._label_num = 0

        self._parse_label(self._raw_label)
            
    def _parse_label(self, label):
        self._label_num = label[0]

        for lab in splitindex(label[1:], 18):
            label = LabelSplit()
            label.parse(lab)
            self._label_list.append(label)

    def append(self, label):
        self._label_num += 1
        self._label_list.append(label)

    def to_array(self):
        result_array = []
        result_array.append(self._label_num)
        self._label_list = sorted(self._label_list, key=lambda x: x._time)

        for lab in self._label_list:
            result_array.extend(lab.encoded_label())

        return result_array


class LabelSplit:
    def __init__(self):
        self._char_label = np.zeros((64, ), dtype="uint32")
        self._time = 0

    def parse(self, raw_label):
        self._time = raw_label[0]
        encoded_text = ""
        
        for i, raw in enumerate(raw_label[2:]):
            encoded_text += format(int.from_bytes(raw, byteorder="big"), 'x')

        for i, byte in enumerate(splitindex(encoded_text, 2)):
            self._char_label[i] = int(byte, base=16)

    def setTime(self, time):
        self._time = time

    def setLabel(self, value):
        self._char_label = np.zeros((64, ), dtype="uint32")
        for i, v in enumerate(value):
            self._char_label[i] = ord(v)

    def encoded_label(self):
        encoded_text = int(self._time).to_bytes(8, 'little')

        for char in self._char_label:
            encoded_text += int(char).to_bytes(1, 'little')

        return np.frombuffer(encoded_text, dtype="I")
